package io.tswf.transmutations.expansion.generator

import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.transforms.validation.checkedmapconstructor.CheckedMapConstructor
import io.tswf.transmutations.common.conflict.ConflictResolver
import io.tswf.transmutations.expansion.meta.ExpansionMetadata
import org.objectweb.asm.tree.*

import static org.objectweb.asm.Opcodes.*

@CompileStatic
class ExpansionInjector {

    static void injectExpansion(ExpansionInjectionContext context) {
        def targetClass = context.targetClass
        def expansionClass = context.expansionMeta.expansionClass
        def expansionClassType = expansionClass.asAsmType()
        def expansionClassSimpleName = expansionClass.javaName.split("\\.").last()

        // Добавляем в класс-цель все нужные интерфейсы
        for (def expansionInterface : context.expansionMeta.interfacesToImplement) {
            targetClass.addInterface(expansionInterface)
        }

        // Используем рандомный постфикс для того, чтобы точно не породить конфликт имен
        def expansionRandomPostfix = Random.randomString(16)

        // Создаем поле, в котором будет храниться экземпляр нашей экспансии
        def expansionField = new FieldNode(ACC_PUBLIC, "_expansion_${expansionClassSimpleName}_${expansionRandomPostfix}", expansionClassType.descriptor, null, null)

        // Создаем метод, который будет инициализировать поле с экспансией
        def expansionInitMethod = new MethodNode(ACC_PUBLIC, "_init_expansion_${expansionClassSimpleName}_${expansionRandomPostfix}", "()V", null, null)

        expansionInitMethod.instructions << new VarInsnNode(ALOAD, 0)
        expansionInitMethod.instructions << new TypeInsnNode(NEW, expansionClassType.internalName)
        expansionInitMethod.instructions << new InsnNode(DUP)
        expansionInitMethod.instructions << new MethodInsnNode(INVOKESPECIAL, expansionClassType.internalName, "<init>", "()V", false)
        expansionInitMethod.instructions << new FieldInsnNode(PUTFIELD, targetClass.name, expansionField.name, expansionField.desc)
        expansionInitMethod.instructions << new InsnNode(RETURN)

        // Добавляем вызов этого метода во все конструкторы класса-цели
        for (def constructor : targetClass.methods.filter { it.name == "<init>" }) {
            def allReturnInsns = constructor.instructions.filter { it.opcode == RETURN }
            for (def returnInsn : allReturnInsns) {
                def invokeInitExpansionInsns = new InsnList()
                invokeInitExpansionInsns << new VarInsnNode(ALOAD, 0)
                invokeInitExpansionInsns << new MethodInsnNode(INVOKEVIRTUAL, targetClass.name, expansionInitMethod.name, expansionInitMethod.desc, false)

                constructor.instructions.insertBefore(returnInsn, invokeInitExpansionInsns)
            }
        }

        // Добавляем новое поле и метод для хранения и инициализации экспансии в класс-цель
        targetClass.fields << expansionField
        targetClass.methods << expansionInitMethod

        // Создаем биндинги для каждого метода экспансии. Биндинги будут загружать инстанс экспансии и вызывать у него точно такой же метод, как и они сами, выступая таким образом в качестве прокси.
        // Это не будет захламлять класс-цель лишним байткодом и избавит от геморроя с ремаппингом кастомных JVM языков.
        for (def expansionMethodMeta : context.expansionMeta.expansionMethodsMetadata) {
            def expansionMethod = expansionMethodMeta.expansionMethod
            def conflictingMethodInTargetClass = targetClass.findSameMethod(expansionMethod)

            if (conflictingMethodInTargetClass != null) {
                def canOverwriteMethod = ConflictResolver.resolveConflict(
                        targetClass: targetClass,
                        conflictingMethod: conflictingMethodInTargetClass,
                        anotherMethodOwner: expansionMethodMeta.expansion.expansionClass,
                        anotherMethod: expansionMethodMeta.expansionMethod,
                        strategyProvider: expansionMethodMeta
                )

                if (!canOverwriteMethod)
                    continue

                targetClass.methods.remove(conflictingMethodInTargetClass)
            }

            def expansionBindingMethod = new MethodNode(expansionMethod.access, expansionMethod.name, expansionMethod.desc, expansionMethod.signature, null)

            // Загружаем нашу экспансию
            expansionBindingMethod.instructions << new VarInsnNode(ALOAD, 0)
            expansionBindingMethod.instructions << new FieldInsnNode(GETFIELD, targetClass.name, expansionField.name, expansionField.desc)

            // Загружаем параметры метода
            def expansionBindingMethodParameterTypes = expansionBindingMethod.parameterTypes
            for (def parameterIndex = 0; parameterIndex < expansionBindingMethodParameterTypes.size(); parameterIndex ++) {
                def expansionBindingMethodParameterType = expansionBindingMethodParameterTypes[parameterIndex]
                expansionBindingMethod.instructions << new VarInsnNode(expansionBindingMethodParameterType.getOpcode(ILOAD), parameterIndex + 1)
            }

            // Вызываем метод экспансии
            expansionBindingMethod.instructions << new MethodInsnNode(INVOKEVIRTUAL, expansionClassType.internalName, expansionMethod.name, expansionMethod.desc, false)

            // Возвращаем его результат
            expansionBindingMethod.instructions << new InsnNode(expansionMethod.returnType.getOpcode(IRETURN))

            // Добавляем полученный метод в класс-цель
            targetClass.methods << expansionBindingMethod
        }
    }
}

@CheckedMapConstructor
@CompileStatic
class ExpansionInjectionContext {
    ClassNode targetClass
    ExpansionMetadata expansionMeta
}
