package io.tswf.transmutations.expansion.reader

import groovy.transform.CompileStatic
import io.tswf.transmutations.common.reader.ClassReferenceReader
import io.tswf.transmutations.common.remapping.TransmutationRemapper
import io.tswf.transmutations.expansion.annotations.Expand
import io.tswf.transmutations.expansion.annotations.Expansion
import io.tswf.transmutations.expansion.annotations.Ignore
import io.tswf.transmutations.expansion.exception.ExpansionMetadataReadingException
import io.tswf.transmutations.expansion.exception.ExpansionMethodMetadataReadingException
import io.tswf.transmutations.expansion.exception.InvalidExpansionClassException
import io.tswf.transmutations.expansion.meta.ExpansionMetadata
import io.tswf.transmutations.expansion.reader.exclusion.IExpansionCustomExclusion
import io.tswf.transmutations.expansion.util.ReservedNames
import org.objectweb.asm.Type
import org.objectweb.asm.tree.AnnotationNode
import org.objectweb.asm.tree.ClassNode

import java.lang.reflect.Modifier
import java.util.stream.Stream

import static org.objectweb.asm.Opcodes.ACC_SYNTHETIC

@CompileStatic
class ExpansionMetadataReader {

    @Lazy
    static List<IExpansionCustomExclusion> customExclusions = ServiceLoader.load(IExpansionCustomExclusion)
            .toList()

    public static final String TARGET_CLASSES_PARAMETER_NAME = "targets"
    public static final String EXCLUDED_ANNOTATIONS_PARAMETER_NAME = "excludeInterfaces"

    static ExpansionMetadata readExpansionMetadata(ClassNode expansionClass) {
        def expandAnnotation = expansionClass.getAnnotationByType(Expand)

        def result = new ExpansionMetadata()
        result.expansionClass = expansionClass

        try {

            // Объект экспансии создается через no-args конструктор, так что надо убедиться в том, что он есть
            expansionClass.methods
                    .find { it.name == "<init>" && it.desc == "()V" }
                    .ensureNotNull { new InvalidExpansionClassException("it has no default constructor", result) }

            def rawTargets = new HashSet<Type>()

            // Читаем цели для расширения
            expandAnnotation.getValueByName(TARGET_CLASSES_PARAMETER_NAME, AnnotationNode[])
                    .tap { annotationNodes ->
                        def targetTypes = annotationNodes
                                .map { ClassReferenceReader.findTypeFromClassRefAnnotation(it) }

                        rawTargets.addAll(targetTypes)
                    }

            def interfacesToExclude = new HashSet<Type>()
            // Читаем аннотации, которые должны быть исключены при инъекции расширения
            expandAnnotation.findValueByName(EXCLUDED_ANNOTATIONS_PARAMETER_NAME, Type[])?.tap {
                interfacesToExclude.addAll(Arrays.asList(it))
            }

            // Проверяем, что пользователь указал хотя бы одну цель
            if (rawTargets.isNullOrEmpty()) {
                throw new InvalidExpansionClassException("no one target provided in @${Expand.simpleName} annotation", result)
            }

            result.targets = rawTargets
                    .mapToCollection({ TransmutationRemapper.mapType(it) }, { new HashSet<Type>() })

            // Формируем коллекцию интейрфейсов, которые надо реализовать в классе-цели
            def rawInterfacesToImplement = Stream.of(expansionClass.interfaces)
                    .filter { it != null }
                    .flatMap { it.stream() }
                    .map { Type.getObjectType(it) }
                    .filter { it !in interfacesToExclude }
                    .filter { !customExclusions.any { excl -> excl.isNeedsToExcludeInterface(result, it)} }
                    .toSet()

            result.interfacesToImplement = rawInterfacesToImplement
                    .mapToCollection({ TransmutationRemapper.mapType(it) }, { new HashSet<Type>() })

            // Читаем методы-расширения, которые надо будет прокинуть в класс-цель совместно с интерфейсами при инъекции

            // Такими методами являются публичные, не-статические, не-синтетические и имеющие имя, не начинающееся с '_' (Последнее сделано для удобства, чтобы не расставлять везде @Ignore)
            def notAnnotatedExpansionMethods = expansionClass.methods
                    .filter { Modifier.isPublic(it.access) && (it.access & ACC_SYNTHETIC) == 0 }
                    .filter { !it.name.startsWith("_") }
                    .filter { it.name !in ReservedNames.CLASS_INTERNAL_METHOD_NAMES }
                    .filter { !customExclusions.any { excl -> excl.isNeedsToExcludeMethod(result, it) } }

            // Еще добавляем для верности все методы, которые отмечены аннотацией @Expansion
            def expansionAnnotatedMethods = expansionClass.methods
                    .filter { it.hasAnnotation(Expansion) }

            // Собираем все в один сет и удаляем из него те методы, которые отмечены аннотацией Ignore. Дубликатов в сете не будет потому, что MethodNode используются для обоих списков из одного и того же объекта ClassNode.
            def allExpansionMethods = (expansionAnnotatedMethods + notAnnotatedExpansionMethods)
                .stream()
                    .filter { !it.hasAnnotation(Ignore) }
                    .toSet()

            // Начинаем читать метадату методов, отмеченных для переноса в качестве прокси в класс-цель и добавляем их в мету экспансии
            for (def expansionMethod : allExpansionMethods) {
                try {
                    def expansionMethodMeta = ExpansionMethodMetadataReader.readExpansionMethodMetadata(result, expansionMethod)
                    result.expansionMethodsMetadata << expansionMethodMeta
                } catch (Throwable e) {
                    throw new ExpansionMethodMetadataReadingException(e, result, expansionMethod)
                }
            }
        } catch (Throwable e) {
            throw new ExpansionMetadataReadingException(e, result)
        }

        return result
    }
}
